<?php
namespace Zeedhi\Tools\TableImporter;

use Zeedhi\Tools\DatabaseReader\DTO;
use Zeedhi\Tools\DatabaseReader\DatabaseReader;
use Zeedhi\Tools\DatabaseReader\Exception as DataBaseReaderException;

class TableImporter {

    /** @var DatabaseReader */
    protected $dataBaseReader;
    /** @var Array */
    protected $data = array();

    function __construct($dataBaseReader, $clean) {
        $this->dataBaseReader = $dataBaseReader;
        $this->clean = $clean;
    }

    /**
     * @param string $tableName
     *
     * @throws \Zeedhi\Tools\Generator\Exception Table not found in database.
     *
     * @return Table
     */
    protected function createTable($tableName) {
        return $this->dataBaseReader->getTable($tableName);
    }

    /**
     * @param string $subModelName
     * @return Submodel
     */
    protected function createSubModel($subModelName) {
        $subModel = new DTO\Submodel();
        $subModel->setName($subModelName);
        return $subModel;
    }

    /**
     * @param Table $table
     * @param string  $subModelName
     */
    protected function assocTableToSubModel(DTO\Table $table, $subModel) {
        $subModel->addTable($table, $table->getName());
    }

    public function getData(){
        return $this->data;
    }

    /**
     * @param Table $table
     *
     * @return Column[] Indexed by name.
     */
    protected function createColumnsForTable(DTO\Table $table) {
        $tableName = $table->getName();
        $columns = $this->dataBaseReader->getColumnsFromTable($tableName);
        foreach ($columns as $columnName => $column) {
            $table->addColumn($column, $columnName);
            $column->setTable($table);
        }
    }

    /**
     * @param Table    $table
     * @param Column[] $columnsByName
     */
    protected function createPrimaryKeyForTable(DTO\Table $table) {
        try {
            $primaryKey = $this->dataBaseReader->getPrimaryKeyFromTable($table->getName());
            $primaryKey->setConstraintType(DTO\Constraint::CONSTRAINT_TYPE_PK);
            $primaryKey->setTable($table);
            $table->addConstraint($primaryKey);
            return $primaryKey;
        } catch (DataBaseReaderException $e) {
            // this table has no pk. so nothing to do here.
            $this->warning("Table {table} has not a primary key.", array('table' => $table->getName()));
        }
    }

    /**
     * @param Table $table
     * @param Column $column
     */
    protected function importSequenceForColumn(DTO\Table $table, DTO\Column $column) {
        $columnName = $column->getName();
        $sequence = $this->dataBaseReader->getSequenceFromTableAndColumn($table->getName(), $columnName);
        if($sequence) {
            $sequence->setColumn($column);
        }
        return $sequence;
    }

    /**
     * @param Table $table
     * @param Column[] $columns
     */
    protected function importSequenceForTable(DTO\Table $table, DTO\Constraints\PrimaryKey $actualPk) {
        $tableName = $table->getName();
        $consColumns = $actualPk->getConsColumns();
        $sequences = [];
        foreach ($consColumns as $consColumn) {
            $column = $consColumn->getColumn();
            if($sequence = $this->importSequenceForColumn($table, $column)) $sequences[$column->getName()] = $sequence;
        }
        return $sequences;
    }

    /**
     * @param Table    $table
     * @param Column[] $columnsByName
     */
    protected function createForeignKeysForTable(DTO\Table $table) {
        $fks = array();
        foreach($this->dataBaseReader->getForeignKeysFromTable($table->getName()) as $foreignKey){
            $foreignKey->setTable($table);
            $fks[] = $foreignKey;
        }
        return $fks;
    }

    protected function createUniqueKeyConstraintForTable(DTO\Table $table){
        $uks = array();
        foreach($this->dataBaseReader->getUniqueConstraintsFromTable($table->getName()) as $uniqueKey){
            $uniqueKey->setTable($table);
            $uks[] = $uniqueKey;
        }
        return $uks;
    }

    /**
     * @param $table
     * @param $subModel
     * @return bool
     */
    protected function importTable($tableName, $subModel) {
        try {
            $table = $this->createTable($tableName);
            $this->createColumnsForTable($table);
            $ukConstraints = $this->createUniqueKeyConstraintForTable($table);
            $this->assocTableToSubModel($table, $subModel);
            return array("TABLE" => $table, "UNIQUE" => $ukConstraints);
        } catch (DataBaseReaderException $e) {
            //table does not exist in database reader connection.
            $this->error($e->getMessage());
            return false;
        }
    }

    protected function importTablesSubModels($tableWithSubModels, $subModels = array(), $tables = array(), $constraints = array(), $sequences = array()){

        foreach (array_unique($tableWithSubModels) as $subModelName) {
            $subModels[$subModelName] = $this->createSubModel($subModelName);
        }

        foreach($tableWithSubModels as $tableName => $subModelName){
            if($table = $this->importTable($tableName, $subModels[$subModelName])){
                $data = $table;
                $tables[$tableName] = $data["TABLE"];
                $constraints["U"][$tableName] = $data["UNIQUE"];
            }
        }

        foreach($tables as $tableName => $table){
            $actualPk = $this->createPrimaryKeyForTable($table);
            $constraints["P"][$tableName][] = $actualPk;
        }

        if(!$this->clean){
            foreach ($tables as $tableName => $table) {
                $constraints["F"][$tableName] = $this->createForeignKeysForTable($table);
            }
        }

        return array(
            "SUBMODELS" => $subModels,
            "TABLES" => $tables,
            "CONSTRAINTS" => $constraints,
        );
    }

    /**
     * Generates tables, constraints and sequences objects.
     *
     * @param string[] $tablesWithSubModels The tables and submodels to generate.
     *
     * @return int The number of tables to generate.
     */
    public function importTablesWithSubModel(array $tableWithSubModels) {
        $this->data = $this->importTablesSubModels($tableWithSubModels);
        foreach ($this->data["TABLES"] as $tableName => $table) {
            $columns = $table->getColumns();
            $actualPk = null;
            foreach ($table->getConstraints() as $constraint) {
                $consColumns = $constraint->getConsColumns();
                foreach ($consColumns as $consColumn) {
                    $consColumn->setColumn($columns[$consColumn->getColumnName()]);
                    $table->addConsColumn($consColumn);
                }

                if($constraint->getConstraintType() == "P"){
                    $actualPk = $constraint;
                }
            }

            if($actualPk){
                $this->data["SEQUENCES"][$tableName] = $this->importSequenceForTable($table, $actualPk);
            }
        }
        return count($this->data["TABLES"]);
    }

    /**
     * @param string $message
     * @param array  $context
     */
    protected function warning($message, $context = array()) {
        echo "[warning] ", str_replace(array_keys($context), $context, $message), PHP_EOL;
    }

    /**
     * @param string $message
     * @param array  $context
     */
    protected function error($message, $context = array()) {
        echo "[critical error] ", str_replace(array_keys($context), $context, $message), PHP_EOL;
    }
}
